
PT_SIZE_4K = 4 * 1024
PT_SIZE_16K = 16 * 1024
PT_SIZE_64K = 64 * 1024
PT_SIZE_1MIB = 1024 * 1024
PT_SIZE_2MIB = 2 * PT_SIZE_1MIB
PT_SIZE_32MIB = 32 * PT_SIZE_1MIB
PT_SIZE_512MIB = 512 * PT_SIZE_1MIB
PT_SIZE_1GIB = 1024 * 1024 * 1024
PT_SIZE_64GIB = 64 * PT_SIZE_1GIB
PT_SIZE_512GIB = 512 * PT_SIZE_1GIB
PT_SIZE_1TB = 1024 * PT_SIZE_1GIB
PT_SIZE_4TB = 4 * PT_SIZE_1TB
PT_SIZE_128TB = 128 * PT_SIZE_1TB
